// Display data using parallel coordinates.  Each dimensional axis is layed

// out in a vertical line, evenly spaced between dimensions.  The top and

// bottom of each line coincide with the maximum and minimum for that

// dimension.  Each data point creates a polyline across these axes.  Labels

// are printed along the top of each dimension, and brush coverage display

// consists of a single shaded region across the drawing area.



#include "CParcoord.H"

#include "XmdvTool.H"

#include "ogl.H"





CParcoord::CParcoord()

{

}





CParcoord::~CParcoord()

{

}





bool CParcoord::ParcoordTestDimensionality()

{

	if ( dataset->dims < 2 )

		return false;

	else 

		return true;

}



// Draw standard vertical uniformly spaced parallel coordinate axes.

void CParcoord::DrawParcoordAxes(grCanvas *canvas)

{

    glLineWidth(1.);

    glShadeModel(GL_SMOOTH);

 //   canvas->SetXorMode(false);



    // Draw the vertical axes

    //

    canvas->SetForeground(interface_color[GRID1]);



    int i;

    for (i=0; i<dataset->dims; i++) {

        double x = XMAP(i);

        canvas->DrawLine(x, 0, x, 1.0);

    }



  canvas->SetForeground(interface_color[TEXT]);



    Vec2 font_size = canvas->GetDataFontSize("W");

//    double max_str_size = dimwidth - font_size.X;



	double max_str_size;



    for (i=0; i<dataset->dims; i++)

    {

		if ( i == dataset->dims-1 )

            max_str_size = screen_distortion->paracoord_interval[i]; 

		else

            max_str_size = screen_distortion->paracoord_interval[i] - font_size.X*5; 

        //max_str_size = screen_distortion->scatter_X_interval[i] - font_size.X; 

        //double x = CFlatParcoord::XMAP(i) + font_size.X;

        double x = screen_distortion->paracoord_axes[i];

        canvas->DrawConfinedString(x,font_size.Y,dataset->names[i], max_str_size);



        // Draw the MAX2 values of the dimension

        char buf[50];

        sprintf(buf, "%.02f", dataset->dim_max[i]);

        canvas->DrawConfinedString(x,-.5*font_size.Y,buf,max_str_size);



        // Draw the MIN2 values of the dimension

        sprintf(buf, "%.02f", dataset->dim_min[i]);

        canvas->DrawConfinedString(x,1.0-font_size.Y,buf,max_str_size);

  

    }

}





// Draw standard vertical uniformly spaced parallel coordinate axes.

void CParcoord::DrawParcoordAxesWithBands(grCanvas *canvas)

{

    glLineWidth(1.);

    glShadeModel(GL_SMOOTH);



    // Draw the vertical axes

    //

    canvas->SetForeground(interface_color[GRID1]);



    int i;

    for (i=0; i<dataset->dims; i++) {

        double x = XMAP(i);

        canvas->DrawLine(x, 0, x, 1.0);

    }



    canvas->SetForeground(interface_color[TEXT]);



    Vec2 font_size = canvas->GetDataFontSize("W");

//    double max_str_size = dimwidth - font_size.X;



	double max_str_size;



    for (i=0; i<dataset->dims; i++)

    {

		max_str_size = screen_distortion->paracoord_interval[i] - font_size.X*5; 

		//max_str_size = screen_distortion->scatter_X_interval[i] - font_size.X; 

        //double x = CFlatParcoord::XMAP(i) + font_size.X;

        double x = screen_distortion->paracoord_axes[i];

//        canvas->DrawConfinedString(x,font_size.Y,dataset->names[i], max_str_size);

        char name[50], name_a[30];

        strcpy(name, dataset->names[i]);

        sprintf(name_a, "(%1d)", dim_reducer->GetEntries(i));

        strcat(name, name_a);

//        canvas->DrawConfinedString(x,font_size.Y,dataset->names[i], max_str_size);

        canvas->DrawConfinedString(x,font_size.Y, name, max_str_size);



        // Draw the MAX2 values of the dimension

        char buf[50];

        sprintf(buf, "%.02f", dataset->dim_max[i]);

        canvas->DrawConfinedString(x,-.5*font_size.Y,buf,max_str_size);



        // Draw the MIN2 values of the dimension

        sprintf(buf, "%.02f", dataset->dim_min[i]);

        canvas->DrawConfinedString(x,1.0-font_size.Y,buf,max_str_size);

    }

}



// Draw standard vertical uniformly spaced parallel coordinate axes.

void CParcoord::DrawParcoordWideAxes(grCanvas *canvas)

{

    glShadeModel(GL_SMOOTH);



    // Draw the vertical axes

    //

    canvas->SetForeground(interface_color[GRID1]);



    int i;

    for (i=0; i<dataset->dims; i++) {

        /* set width of the axis */

        if ( dim_reducer->IsLeafDim(i) )

            glLineWidth(1.);

        else 

        {

            double dim_disagreement = dim_reducer->GetDimDisagreement(i);

            glLineWidth((float)(1.+ dim_disagreement * 10.));

        }

        double x = XMAP(i);

        canvas->DrawLine(x, 0, x, 1.0);

    }



    canvas->SetForeground(interface_color[TEXT]);



    Vec2 font_size = canvas->GetDataFontSize("W");

//    double max_str_size = dimwidth - font_size.X;



	double max_str_size;



    for (i=0; i<dataset->dims; i++)

    {

		max_str_size = screen_distortion->paracoord_interval[i] - font_size.X*5; 

		//max_str_size = screen_distortion->scatter_X_interval[i] - font_size.X; 

        //double x = CFlatParcoord::XMAP(i) + font_size.X;

        double x = screen_distortion->paracoord_axes[i];

        char name[50], name_a[30];

        strcpy(name, dataset->names[i]);

        sprintf(name_a, "(%1d)", dim_reducer->GetEntries(i));

        strcat(name, name_a);

//        canvas->DrawConfinedString(x,font_size.Y,dataset->names[i], max_str_size);

        canvas->DrawConfinedString(x,font_size.Y, name, max_str_size);



        // Draw the MAX2 values of the dimension

        char buf[50];

        sprintf(buf, "%.02f", dataset->dim_max[i]);

        canvas->DrawConfinedString(x,-.5*font_size.Y,buf,max_str_size);



        // Draw the MIN2 values of the dimension

        sprintf(buf, "%.02f", dataset->dim_min[i]);

        canvas->DrawConfinedString(x,1.0-font_size.Y,buf,max_str_size);

    }

}



// Draw standard vertical uniformly spaced parallel coordinate axes, with two axes to show disagree of dim cluster.

void CParcoord::DrawParcoordAxesWithTwoAxes(grCanvas *canvas)

{

    glLineWidth(1.);

    glShadeModel(GL_SMOOTH);



    // Draw the mean vertical axes

    //

    canvas->SetForeground(interface_color[GRID1]);

    int i;

    for (i=0; i<dataset->dims; i++) 

    {

        double x = XMAP(i);

        canvas->DrawLine(x, 0, x, 1.0);

    }



    // Draw the min and max vertical axes

    //

    for (i=0; i<dataset->dims; i++) 

    {

        if ( !dim_reducer->IsLeafDim(i) )

        {

            canvas->SetForeground(interface_color[GRID2]);

            double x = XMINMAP(i);

            canvas->DrawLine(x, 0, x, 1.0);



            canvas->SetForeground(interface_color[GRID3]);

            x = XMAXMAP(i);

            canvas->DrawLine(x, 0, x, 1.0);

        }

    }



    canvas->SetForeground(interface_color[TEXT]);



    Vec2 font_size = canvas->GetDataFontSize("W");



	double max_str_size;



    for (i=0; i<dataset->dims; i++)

    {

		max_str_size = screen_distortion->paracoord_interval[i] - font_size.X*5; 

        double x = screen_distortion->paracoord_axes[i];

//        canvas->DrawConfinedString(x,font_size.Y,dataset->names[i], max_str_size);

        char name[50], name_a[30];

        strcpy(name, dataset->names[i]);

        sprintf(name_a, "(%1d)", dim_reducer->GetEntries(i));

        strcat(name, name_a);

//        canvas->DrawConfinedString(x,font_size.Y,dataset->names[i], max_str_size);

        canvas->DrawConfinedString(x,font_size.Y, name, max_str_size);



        // Draw the MAX2 values of the dimension

        char buf[50];

        sprintf(buf, "%.02f", dataset->dim_max[i]);

        canvas->DrawConfinedString(x,-.5*font_size.Y,buf,max_str_size);



        // Draw the MIN2 values of the dimension

        sprintf(buf, "%.02f", dataset->dim_min[i]);

        canvas->DrawConfinedString(x,1.0-font_size.Y,buf,max_str_size);

    }

}





void CParcoord::InputParacoordFindAxesBetween(grCanvas *canvas, double X, double Y, int &dim)

{

    if ( Y < 0 || Y > 1.0 )

    {

        dim = -1;

    }

    else

    {

        int i = -1;

        while ( i < dataset->dims-1 && X > screen_distortion->paracoord_axes[i+1] )

            i++;



        dim = i;

    }

    return;

}





void CParcoord::InputParacoordFindAxesNear(grCanvas *canvas, double X, double Y, int &dim)

{

    if ( Y < 0. || Y > 1.0 )

    {

        dim = -1;

        return;

    }

    else

    {

        if ( X < 0. || X > 1.0 )

        {

            dim = -1;

            return;

        }



        for ( int i = 0; i < dataset->dims-1; i++ )

        {

            if ( X < ( screen_distortion->paracoord_axes[i] + screen_distortion->paracoord_interval[i]/2 ) )

            {

                dim = i;

                return;

            }

        }



        dim = dataset->dims-1;

        return;

    }

}





//this func is used by InputParcoordMouseMotion

void CParcoord::ParcoordCaption(grCanvas *canvas, int x, int y)

{

    Vec2 p = canvas->MapScreenToData( Vec2(x,y) );



//    int datax = int(p.X * dataset->dims);

//    datax = CLAMP(datax,0,dataset->dims-1);

    int datax;

    

    InputParacoordFindAxesNear(canvas, p.X, p.Y, datax);



    datax = CLAMP(datax,0,dataset->dims-1);

    

    double datay = YINVMAP(datax,p.Y);



    char caption[100];



    if (dataset->nominal_fields[datax] != 0) {

		int value = (int)(datay+0.5);

        const char *string = dataset->nominal_fields[datax]->MapValueToString(value);

        sprintf(caption,"%s=%s",dataset->names[datax],string);

    } else {

        sprintf(caption,"%s=%.2lf",dataset->names[datax],datay);

    }



    appShowCanvasMessage(Togl_Interp(canvas->GetWidget()),caption, canvas->name);

}





void CParcoord::InputParacoordScaleAxes(grCanvas *canvas, int dim, double delta)

{

    if ( dim < 0 || dim > dataset->dims )

        return;



    screen_distortion->ParacoordScaleAxes(dim, delta);

    return;

}





double CParcoord::XMAP(int I)

{ 

    return screen_distortion->paracoord_axes[I];

}





double CParcoord::XMINMAP(int I)

{ 

    return screen_distortion->paracoord_axes[I]-0.04;

}



double CParcoord::XMAXMAP(int I)

{ 

    return screen_distortion->paracoord_axes[I]+0.04;

}





double CParcoord::YMAP(int X, double Y)

{

    return 1.0-(Y - dataset->dim_min[X]) / (dataset->dim_max[X] - dataset->dim_min[X]);

}





//it is for the uniformed dim: 0-1

double CParcoord::YMAP(double Y)

{

    return 1.0-Y;

}





double CParcoord::YINVMAP(int X, double Y)

{

    return ((1.0-(Y)) * (dataset->dim_max[X] - dataset->dim_min[X])) + dataset->dim_min[X];

}



