
#include "cluster.H"
#include "XmdvTool.H"
#include "brush.H"
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "CHierNav.H"
#include "CHierHist.H"
#include "CHierTreemap.H"
#include "OnlineDataClustering.H"

/* SID */
#include "./backend_fake/db_buffer.h"


ClusterTree *btree = 0;


bool ClusterTree::ReadClusterTreeFile(const char* fname)
{
    int k, i;
    assert(dataset);
   
    //test if file exists. If not, online generate one.
    FILE *fptr = fopen(fname,"rb");
    if (fptr==NULL) {
        OnlineClusterOkcFileToCgFile(dataset->datafile, 100, true);
        fptr = fopen(fname,"rb");
        if (fptr==NULL) 
            return false;
    }

    //test if binary file.
    unsigned int magic_code;
    fread(&magic_code, sizeof(unsigned int), 1, fptr);
    
    fclose(fptr);

    if ( magic_code == 0xcecfcecf )
    {
        return ReadBinaryClusterTreeFile(fname);
    }
    
    fptr = fopen(fname,"r");

    clusterfile = strdup(fname);

    // Get file size
    int num_nodes;
    if (fscanf(fptr, "%d %d\n", &num_nodes, &dimension)<=0)
    {
        printf("Error reading the first line of the file %s\n", fname);
        printf("Format of the first line: #nodes #dimension\n");
        exit(-1);
    }
    assert(dimension == dataset->all_dims);

    // Allocate space for all nodes
    //
    nodes.SetLen(num_nodes);

//  selected.SetLen(num_nodes/2+1);

    printf("Number of cluster nodes=%d\n", num_nodes);


    for (i=0; i<num_nodes; i++) {
        nodes[i].sx.SetLen(dimension);
        nodes[i].mean.SetLen(dimension);
        nodes[i].max.SetLen(dimension);
        nodes[i].min.SetLen(dimension);
      //  nodes[i].all_mean.SetLen(dimension);
        nodes[i].num_children = 0;
        nodes[i].entries = 0;
    }

    for (k=0; k<num_nodes; k++ )
        for (i=0;i<dimension; i++)
            nodes[k].sx[i] = 0.0;

// SID 

    double max, min;
    for (k=0; k<num_nodes; k++)
    {
        int id;
        fscanf(fptr, "%d %d %d", &id, &nodes[k].parent, &nodes[k].entries);

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &(nodes[k].sx[i]));

            // Compute the mean
            if ( nodes[k].entries > 0) {

                double sum_dim = nodes[k].sx[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
                                 (nodes[k].entries * dataset->cg_dim_min[i]);

                nodes[k].mean[i] = sum_dim/nodes[k].entries;
            }

        }


        fscanf(fptr, "%lf ", &nodes[k].sqrt_radius);

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &max);
			nodes[k].max[i] = max * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &min);
			nodes[k].min[i] = min * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        fscanf(fptr, "\n");

        if ( nodes[k].parent >= 0 ) {
            nodes[nodes[k].parent].child.InsertTail(k);
            nodes[nodes[k].parent].num_children++;
        }

    }

    fclose(fptr);

    // Init the min/max extents
    //
    min_extents.SetLen(dataset->all_dims);
    max_extents.SetLen(dataset->all_dims);

    for (int d=0; d<dataset->all_dims; d++) {
        min_extents[d] = HUGE_VAL;
        max_extents[d] = -HUGE_VAL;
    }

    treemap_limit_number = (int) (nodes[0].entries * TREEMAP_LIMIT);
    if ( treemap_limit_number < 1 )
        treemap_limit_number = 1;

    return true;
}



bool ClusterTree::ReadBinaryClusterTreeFile(const char* fname)
{
    int k, i;
    assert(dataset);
   
    FILE *fptr = fopen(fname,"rb");
    if (fptr==NULL) 
        return false;

    unsigned int magic_code;
    fread(&magic_code, sizeof(unsigned int), 1, fptr);

    clusterfile = strdup(fname);

    // Get file size
    int num_nodes;
    if ( fread(&num_nodes, sizeof(int), 1, fptr) != 1 || fread(&dimension, sizeof(int), 1, fptr) != 1 ) 
    {
        printf("Format of the first line: #nodes #dimension\n");
        exit(-1);
    }
    assert(dimension == dataset->all_dims);

    // Allocate space for all nodes
    //
    nodes.SetLen(num_nodes);

//  selected.SetLen(num_nodes/2+1);

    printf("Number of cluster nodes=%d\n", num_nodes);


    for (i=0; i<num_nodes; i++) {
        nodes[i].sx.SetLen(dimension);
        nodes[i].mean.SetLen(dimension);
        nodes[i].max.SetLen(dimension);
        nodes[i].min.SetLen(dimension);
      //  nodes[i].all_mean.SetLen(dimension);
        nodes[i].num_children = 0;
        nodes[i].entries = 0;
    }

    for (k=0; k<num_nodes; k++ )
        for (i=0;i<dimension; i++)
            nodes[k].sx[i] = 0.0;

// SID 

    double * buffer;
    buffer = new double[dimension];
    for (k=0; k<num_nodes; k++)
    {
        int id;
        fread(&id, sizeof(int), 1, fptr);
        fread(&nodes[k].parent, sizeof(int), 1, fptr);
        fread(&nodes[k].entries, sizeof(int), 1, fptr);


        fread(buffer, sizeof(double), dimension, fptr);
        for (i=0; i<dimension; i++)
		{
              nodes[k].sx[i] = buffer[i];
//            fread(&(nodes[k].sx[i]), sizeof(double), 1, fptr);
//            fscanf(fptr, "%lf ", &(nodes[k].sx[i]));

            // Compute the mean
            if ( nodes[k].entries > 0) {

//                double sum_dim = nodes[k].sx[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
//                                 (nodes[k].entries * dataset->cg_dim_min[i]);
                double sum_dim = buffer[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
                                 (nodes[k].entries * dataset->cg_dim_min[i]);

                nodes[k].mean[i] = sum_dim/nodes[k].entries;
            }

        }


        fread(&(nodes[k].sqrt_radius), sizeof(double), 1, fptr);
//        fscanf(fptr, "%lf ", &nodes[k].sqrt_radius);

        fread(buffer, sizeof(double), dimension, fptr);
        for (i=0; i<dimension; i++)
		{
//            fread(&max, sizeof(double), 1, fptr);
//            fscanf(fptr, "%lf ", &max);
			nodes[k].max[i] = buffer[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        fread(buffer, sizeof(double), dimension, fptr);
        for (i=0; i<dimension; i++)
		{
//            fread(&min, sizeof(double), 1, fptr);
//            fscanf(fptr, "%lf ", &min);
			nodes[k].min[i] = buffer[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

//        fscanf(fptr, "\n");

        if ( nodes[k].parent >= 0 ) {
            nodes[nodes[k].parent].child.InsertTail(k);
            nodes[nodes[k].parent].num_children++;
        }

    }

    fclose(fptr);
    delete buffer;

    // Init the min/max extents
    //
    min_extents.SetLen(dataset->all_dims);
    max_extents.SetLen(dataset->all_dims);

    for (int d=0; d<dataset->all_dims; d++) {
        min_extents[d] = HUGE_VAL;
        max_extents[d] = -HUGE_VAL;
    }

    treemap_limit_number = (int) (nodes[0].entries * TREEMAP_LIMIT);
    if ( treemap_limit_number < 1 )
        treemap_limit_number = 1;

    return true;
}


/*
bool ClusterTree::ReadClusterTreeFile(const char* fname)
{
    int k, i;
    assert(dataset);

    FILE *fptr = fopen(fname,"r");
    if (fptr==NULL) {
        OnlineClusterOkcFileToCgFile(dataset->datafile, 100);
        fptr = fopen(fname,"r");
        if (fptr==NULL) 
            return false;
    }

    clusterfile = strdup(fname);

    // Get file size
    int num_nodes;
    if (fscanf(fptr, "%d %d\n", &num_nodes, &dimension)<=0)
    {
        printf("Error reading the first line of the file %s\n", fname);
        printf("Format of the first line: #nodes #dimension\n");
        exit(-1);
    }
    assert(dimension == dataset->all_dims);

    // Allocate space for all nodes
    //
    nodes.SetLen(num_nodes);

//  selected.SetLen(num_nodes/2+1);

    printf("Number of cluster nodes=%d\n", num_nodes);


    for (i=0; i<num_nodes; i++) {
        nodes[i].sx.SetLen(dimension);
        nodes[i].mean.SetLen(dimension);
        nodes[i].max.SetLen(dimension);
        nodes[i].min.SetLen(dimension);
      //  nodes[i].all_mean.SetLen(dimension);
        nodes[i].num_children = 0;
        nodes[i].entries = 0;
    }

    for (k=0; k<num_nodes; k++ )
        for (i=0;i<dimension; i++)
            nodes[k].sx[i] = 0.0;

// SID 

    double max, min;
    for (k=0; k<num_nodes; k++)
    {
        int id;
        fscanf(fptr, "%d %d %d", &id, &nodes[k].parent, &nodes[k].entries);

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &(nodes[k].sx[i]));

            // Compute the mean
            if ( nodes[k].entries > 0) {

                double sum_dim = nodes[k].sx[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
                                 (nodes[k].entries * dataset->cg_dim_min[i]);

                nodes[k].mean[i] = sum_dim/nodes[k].entries;
            }

        }


        fscanf(fptr, "%lf ", &nodes[k].sqrt_radius);

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &max);
			nodes[k].max[i] = max * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &min);
			nodes[k].min[i] = min * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        fscanf(fptr, "\n");

        if ( nodes[k].parent >= 0 ) {
            nodes[nodes[k].parent].child.InsertTail(k);
            nodes[nodes[k].parent].num_children++;
        }

    }

    // Init the min/max extents
    //
    min_extents.SetLen(dataset->all_dims);
    max_extents.SetLen(dataset->all_dims);

    for (int d=0; d<dataset->all_dims; d++) {
        min_extents[d] = HUGE_VAL;
        max_extents[d] = -HUGE_VAL;
    }

    treemap_limit_number = (int) (nodes[0].entries * TREEMAP_LIMIT);
    if ( treemap_limit_number < 1 )
        treemap_limit_number = 1;

    return true;
}
*/

bool ClusterTree::ReadClusterTreeFileWithDatabase(const char* fname)
{
    int k, i;
    assert(dataset);

    FILE *fptr = fopen(fname,"r");
    if (fptr==NULL) {
        printf("Cluster file \"%s\" does not exist. Program terminated.\n", fname);
//        abort();
        return false;
    }

    clusterfile = strdup(fname);

    // Get file size
    int num_nodes;
    if (fscanf(fptr, "%d %d\n", &num_nodes, &dimension)<=0)
    {
        printf("Error reading the first line of the file %s\n", fname);
        printf("Format of the first line: #nodes #dimension\n");
        exit(-1);
    }
    assert(dimension == dataset->all_dims);

    // Allocate space for all nodes
    //
    nodes.SetLen(num_nodes);

//  selected.SetLen(num_nodes/2+1);

    printf("Number of cluster nodes=%d\n", num_nodes);


    for (i=0; i<num_nodes; i++) {
        nodes[i].sx.SetLen(dimension);
        nodes[i].mean.SetLen(dimension);
      //  nodes[i].max.SetLen(dimension);
      //  nodes[i].min.SetLen(dimension);
        nodes[i].all_mean.SetLen(dimension);
        nodes[i].num_children = 0;
        nodes[i].entries = 0;
    }

/*    for (k=0; k<num_nodes; k++ )
        for (i=0;i<dimension; i++)
            nodes[k].sx[i] = 0.0;
*/
/* SID */

    double max, min, temp;
    for (k=0; k<num_nodes; k++)
    {
        int id;
        fscanf(fptr, "%d %d %d", &id, &nodes[k].parent, &nodes[k].entries);

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &temp);

            // Compute the mean
            if ( nodes[k].entries > 0) {

                double sum_dim = nodes[k].sx[i] * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) +
                                 (nodes[k].entries * dataset->cg_dim_min[i]);

                nodes[k].mean[i] = sum_dim/nodes[k].entries;
            }

        }


        fscanf(fptr, "%lf ", &nodes[k].sqrt_radius);

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &max);
//			nodes[k].max[i] = max * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        for (i=0; i<dimension; i++)
		{
            fscanf(fptr, "%lf ", &min);
//			nodes[k].min[i] = min * (dataset->cg_dim_max[i]-dataset->cg_dim_min[i]) + dataset->cg_dim_min[i];
		}

        fscanf(fptr, "\n");

        if ( nodes[k].parent >= 0 ) {
            nodes[nodes[k].parent].child.InsertTail(k);
            nodes[nodes[k].parent].num_children++;
        }

    }

    // Init the min/max extents
    //
    min_extents.SetLen(dataset->all_dims);
    max_extents.SetLen(dataset->all_dims);

    for (int d=0; d<dataset->all_dims; d++) {
        min_extents[d] = HUGE_VAL;
        max_extents[d] = -HUGE_VAL;
    }

    treemap_limit_number = (int) (nodes[0].entries * TREEMAP_LIMIT);
    if ( treemap_limit_number < 1 )
        treemap_limit_number = 1;

    return true;
}

//
// To mark the highlighted clusters based on the
// brush movement in other modes
//
void hierMarkBrushedClustersByExtents(Cluster* node)
{
    BITCLEAR(node->flags,IS_BRUSHED);
    BITCLEAR(node->flags,IS_BRUSHED_ANY);

    for (int i=0; i<node->num_children; i++)
        hierMarkBrushedClustersByExtents(&(btree->nodes[node->child[i]]));

    // Reached the leaf node?
    //
    if (node->num_children == 0) {
        int in=0;
        for (int k=0; k<dataset->dims; k++) {
            if (node->mean[k] >= btree->min_extents[k] && node->mean[k] <= btree->max_extents[k]) {
                in++;
            }
        }

        if (in == dataset->dims) {
            BITSET(node->flags,IS_BRUSHED);
            BITSET(node->flags,IS_BRUSHED_ANY);
        }

        return;
    }

    // Propagate markers up to parents
    //
    int brushed=0;
    bool brushed_any = false;
    for (int j=0; j<node->num_children; j++) {
        if (BITTEST(btree->nodes[node->child[j]].flags,IS_BRUSHED)) {
            brushed++;
        }
        if (BITTEST(btree->nodes[node->child[j]].flags,IS_BRUSHED_ANY)) {
            brushed_any = true;
        }
    }

	if ( BrushSemantic == 0 ) /* ALL */
	{
		// All my children are marked, mark myself too
		//
		if (brushed == node->num_children)
		{
			BITSET(node->flags,IS_BRUSHED);
		}
	}
	else if ( BrushSemantic == 1 ) /* ANY */
	{
		// Any my children is marked, mark myself too
		//
		if ( brushed_any == true ) 
		{
			BITSET(node->flags,IS_BRUSHED);
		}
	}

	if ( brushed_any == true ) {
		BITSET(node->flags,IS_BRUSHED_ANY);
	}
}

void hierMarkBrushedClustersByColor(Cluster* node, double L, double R)
{
    BITCLEAR(node->flags,IS_BRUSHED);
    BITCLEAR(node->flags,IS_BRUSHED_ANY);

    for (int i=0; i<node->num_children; i++) {
        hierMarkBrushedClustersByColor(&(btree->nodes[node->child[i]]), L, R);
    }

    // Reached the leaf node?
    //
    if (node->num_children == 0)
    {
        if (node->color>=L && node->color<=R)
        {
            BITSET(node->flags,IS_BRUSHED);
            BITSET(node->flags,IS_BRUSHED_ANY);

            // Setup min/max extents
            //
            for (int d=0; d<dataset->dims; d++) {
                if (btree->min_extents[d]>node->mean[d]) btree->min_extents[d]=node->mean[d];
                if (btree->max_extents[d]<node->mean[d]) btree->max_extents[d]=node->mean[d];
            }
        }

        return;
    }

    // Propagate markers up to parents
    //
    int brushed=0;
    bool brushed_any = false;
    for (int j=0; j<node->num_children; j++) {
        if (BITTEST(btree->nodes[node->child[j]].flags,IS_BRUSHED)) {
            brushed++;
        }
        if (BITTEST(btree->nodes[node->child[j]].flags,IS_BRUSHED_ANY)) {
            brushed_any = true;
        }
    }

	if ( BrushSemantic == 0 ) /* ALL */
	{
		// All my children are marked, mark myself too
		//
		if (brushed == node->num_children)
		{
			BITSET(node->flags,IS_BRUSHED);
		}
	}
	else if ( BrushSemantic == 1 ) /* ANY */
	{
		// Any my children is marked, mark myself too
		//
		if ( brushed_any == true ) 
		{
			BITSET(node->flags,IS_BRUSHED);
		}
	}

	if ( brushed_any == true ) {
		BITSET(node->flags,IS_BRUSHED_ANY);
	}
}

void hierModifyBrushParametersByExtents()
{
    for (int i=0; i<dataset->dims; i++) {
        curr_brush->size[i] = btree->max_extents[i] - btree->min_extents[i];
        curr_brush->pos[i] = btree->min_extents[i] + (curr_brush->size[i]/2.);
    }
}

void hierModifyExtentsByBrushParameters()
{
    for (int i=0; i<dataset->dims; i++) {
        btree->min_extents[i] = curr_brush->Min(i);
        btree->max_extents[i] = curr_brush->Max(i);
    }
}

void hierInitClusterColors(Cluster* node, double L, double R)
{
    node->color = .5*(L+R);

    /*SID*/ if (!node->num_children) return;

    double k = (R-L)/(double)(node->num_children);

    for (int i=0; i<node->num_children; i++) {
        hierInitClusterColors(&btree->nodes[node->child[i]], L+i*k, L+(i+1)*k);
    }
}

void hierInitClusterColorsWithDistortion(Cluster* node, double L, double R)
{
    node->color = .5*(L+R);

    double color_l=L;
    double color_r=L;

    for (int i=0; i<node->num_children; i++)
    {
        double k = (R-L) * btree->nodes[node->child[i]].entries/(double)node->entries;
        color_l = color_r;
        color_r = color_l + k;
        hierInitClusterColors(&btree->nodes[node->child[i]], color_l, color_r);
    }
}

void hierDeleteClusterTree()
{
    delete btree;
    btree = 0;
}

void hierInitClusterTree()
{
    if (btree) return;

    btree = new ClusterTree;

    int N = strlen(dataset->datafile);
//    char str[N+1];
//jing
	char* str = new char[N+1];
//jing
    strcpy(str,dataset->datafile);
//jing    strcpy(&str[N-4],".cf");
    strcpy(&str[N-4],".cg");

//jing
	if ( hier_info->database )
	{
//jing for future changes		if (!btree->ReadClusterTreeFileWithDatabase(str)) 
		if (!btree->ReadClusterTreeFile(str)) 
		{
			delete btree;
			btree=0;
			return;
		}
	}
	else
	{
		if (!btree->ReadClusterTreeFile(str)) 
		{
			delete btree;
			btree=0;
			return;
		}
	}

    hierInitClusterColors(&btree->nodes[0], 0, 1);
    hierClearNodeFlags();
    hierModifyExtentsByBrushParameters();
    hierMarkBrushedClustersByExtents(&btree->nodes[0]);

    if ( hier_info->database )
    {
        /* SID */
        init_memory(dataset->datafile);
    }

    cfHistoInit();
    cfNavInit();
    cfTreeMapInit();

    double root_radius = btree->nodes[0].sqrt_radius;
    hierMarkContourNodes( root_radius*hier_info->cluster_radius, IS_CONTOUR0 );
    hierMarkContourNodes( root_radius*hier_info->brushed_radius, IS_CONTOUR1 );
	//jing
	SetupBrushParameters();
}

void hierClearNodeFlags(int bits)
{
    for (int i=0;i<btree->nodes.Len();i++) {
        Cluster &c = btree->nodes[i];
        BITCLEAR(c.flags,bits);
    }
}

void hierMarkContourNodes(double contour_radius, int FLAG)
{
    // A node c is on contour if:
    // c.parent.radius > contour_radius AND
    //  (a)  c.radius <= contour_radius, OR
    //  (b)  c.radius > contour_radius AND c is a leaf
    //
    for (int i=0;i<btree->nodes.Len();i++)
    {
        Cluster &c = btree->nodes[i];

        BITCLEAR(c.flags,FLAG);

        double par_radius = HUGE_VAL;
        if (c.parent>=0) par_radius = btree->nodes[c.parent].sqrt_radius;

        if (par_radius > contour_radius)
        {
            if (c.sqrt_radius <= contour_radius || c.num_children==0)
            {
                assert(c.color<=1.0 && c.color>=0);
                BITSET(c.flags,FLAG);
            }
        }
    }
}

void hierInitExtents()
{
    for (int d=0; d<dataset->all_dims; d++) {
        btree->min_extents[d] = HUGE_VAL;
        btree->max_extents[d] = -HUGE_VAL;
    }
}


void hierSetMeanOnOff()
{
/*	btree->ReadClusterTreeFile2();
    for (int b=0;b<btree->nodes.Len();b++)
    {
        Cluster &c = btree->nodes[b];
        int dims=0;
        for (int i=0;i<dataset->all_dims;i++)
            if (dataset->on[i]) {
                c.mean[dims] = c.mean[i];
                c.max[dims] = c.max[i];
                c.min[dims] = c.min[i];
                dims++;
            }
    }
*/
}


void hierReorderDimension()
{
/*	btree->ReadClusterTreeFile3();
    for (int b=0;b<btree->nodes.Len();b++)
    {
        Cluster &c = btree->nodes[b];
        int dims=0;
        for (int i=0;i<dataset->all_dims;i++)
            if (dataset->on[i]) {
                c.mean[dims] = c.mean[i];
                c.max[dims] = c.max[i];
                c.min[dims] = c.min[i];
                dims++;
            }
    }
*/
}


